// This file is part of FreeFinesse, a double-dummy analyzer (c) Ed Colley, 2006
// It is made available under the GPL; see the file COPYING for details

//
//  Implementation of the analyzer
//

#include "stdafx.h"

// This is the search function for the start of a trick
bool analyzer::search_pl0(uint tricktarget, player_t pl, uint64& rwmask, card_t* winmove, const rankequiv_t& rankequiv)
{
	// Check for trivialities
	if (tricktarget <= 0) return true; 
	if (tricktarget >= 1 + state.tricksLeft()) return false;
	
	// The last trick is dealt with specially (and not cached)
	if (state.tricksLeft() == 1) {
		return search_t13(pl, rwmask);
	}

	// Check the cache
	int cr = cache->check(state, pl, tricktarget, rwmask);
	if (cr != 0) return (cr > 0);
		
	// If none of those applied, we need to search. Start by enumerating possible moves
	// Generate unique moves per suit, keeping track of equivalent alternatives
	card_t suitmoves[4][13];
	uint64 suitrankequivs[4][13];
	int suitmovecount[4];
	for (int s = 0; s < 4; s++) {
		suitmovecount[s] = 0;
		uint64 suit = state.mPlayerHand[pl] & suitmask[s];
		card_t lastmove = -1;
		while (suit) {
			uint64 mcard = lsb(suit);
			suit ^= mcard;
			card_t thismove = bitindex(mcard);
			if ((lastmove == -1) || (rankequiv.nexthigher[lastmove] != thismove)) {
				suitmoves[s][suitmovecount[s]] = thismove;
				suitrankequivs[s][suitmovecount[s]] = mcard;
				suitmovecount[s]++;
			} else {
				suitrankequivs[s][suitmovecount[s]-1] |= mcard;
			}
			lastmove = thismove;
		}
	}
	
	// Assemble moves in preferred order
	card_t moves[13];
	uint64 equivalents[13];
	int movecount = 0;

	// Highest card in each suit
	for (int s = 0; s < 4; s++)	if (suitmovecount[s] > 0) {
		moves[movecount] = suitmoves[s][suitmovecount[s]-1];		
		equivalents[movecount] = suitrankequivs[s][suitmovecount[s]-1];
		movecount++;
	}
		
	// Lowest card in each suit	
	for (int s = 0; s < 4; s++)	if (suitmovecount[s] > 1) {
		moves[movecount] = suitmoves[s][0];							
		equivalents[movecount] = suitrankequivs[s][0];
		movecount++;
	}
		
	// Interior cards
	for (int s = 0; s < 4; s++)	if (suitmovecount[s] > 2) for (int i = 1; i < suitmovecount[s]-1; i++) {
		moves[movecount] = suitmoves[s][i];							
		equivalents[movecount] = suitrankequivs[s][i];
		movecount++;
	}
		
	// Check each move in turn
	uint64 failmask = 0;					// winning cards in failing lines
	int oppotarget = 1 + state.tricksLeft() - tricktarget;
	for (int i = 0; i < movecount; i++) {
		bool thisPlayWorks = false;
		uint64 thismask = 0;
		state.play(moves[i], pl);
		trickstate_t trickstate(pl, moves[i]);
		thisPlayWorks = !search_pl1(oppotarget, nextpl(pl), thismask, 0, rankequiv, trickstate);
		state.unplay();
		if ((thismask & equivalents[i]) != 0) thismask |= sameRankOrHigher[moves[i]];
		if (thisPlayWorks) {
			if (winmove) *winmove = state.cardsPlayed[state.nCardsPlayed];
			cache->update_hit(state, pl, thismask, tricktarget);
			rwmask |= thismask;
			return true;
		} else {
			failmask |= thismask;
		}
	}

	// If we get here, we didn't find a winning move
	cache->update_miss(state, pl, failmask, tricktarget);
	rwmask |= failmask;
	return false;
}

// This is the search function for trick 13; returns true if leader wins it
bool analyzer::search_t13(player_t pl, uint64& rwmask)
{
	trickstate_t trickstate(pl, bitindex(state.mPlayerHand[pl]));	   player_t thispl = nextpl(pl);
	trickstate.play(thispl, bitindex(state.mPlayerHand[thispl]), deal.trumps);	thispl = nextpl(thispl);
	trickstate.play(thispl, bitindex(state.mPlayerHand[thispl]), deal.trumps);	thispl = nextpl(thispl);
	trickstate.play(thispl, bitindex(state.mPlayerHand[thispl]), deal.trumps);			
	if (trickstate.ranktrick) rwmask |= sameRankOrHigher[trickstate.winningcard];
	return (partnership(pl) == partnership(trickstate.winner));
}

// This is the search function for the second player to play to the trick
bool analyzer::search_pl1(uint tricktarget, player_t pl, uint64& rwmask, card_t* winmove, const rankequiv_t& rankequiv, const trickstate_t& trickstate)
{
	// Enumerate possible moves
	card_t moves[13];
	uint64 equivalents[13];
	int movecount = state.generateMoves(pl, trickstate, rankequiv, moves, equivalents);
	
	// Prepare for searching
	int oppotarget = 1 + state.tricksLeft() - tricktarget;	
	uint64 failmask = 0;					// winning cards in failing lines
	
	// Try each move in turn
	for (int i = 0; i < movecount; i++) {
		bool thisPlayWorks = false;
		uint64 thismask = 0;
		state.play(moves[i], pl);
		trickstate_t trickstate_next = trickstate;
		trickstate_next.play(pl, moves[i], deal.trumps);
		thisPlayWorks = !search_pl2(oppotarget, nextpl(pl), thismask, 0, rankequiv, trickstate_next);
		state.unplay();
		if ((thismask & equivalents[i]) != 0) thismask |= sameRankOrHigher[moves[i]];
		if (thisPlayWorks) {
			if (winmove) *winmove = state.cardsPlayed[state.nCardsPlayed];
			rwmask |= thismask;
			return true;
		} else {
			failmask |= thismask;
		}
	}
	
	// If we get here, we didn't find a winning move
	rwmask |= failmask;
	return false;
}

// This is the search function for the third player to play to the trick
bool analyzer::search_pl2(uint tricktarget, player_t pl, uint64& rwmask, card_t* winmove, const rankequiv_t& rankequiv, const trickstate_t& trickstate)
{
	// Enumerate possible moves
	card_t moves[13];
	uint64 equivalents[13];
	int movecount = state.generateMoves(pl, trickstate, rankequiv, moves, equivalents);
	
	// Prepare for searching
	int oppotarget = 1 + state.tricksLeft() - tricktarget;	
	uint64 failmask = 0;					// winning cards in failing lines
	
	// Try each move in turn
	for (int i = 0; i < movecount; i++) {
		bool thisPlayWorks = false;
		uint64 thismask = 0;
		state.play(moves[i], pl);
		trickstate_t trickstate_next = trickstate;
		trickstate_next.play(pl, moves[i], deal.trumps);		
		thisPlayWorks = !search_pl3(oppotarget, nextpl(pl), thismask, 0, rankequiv, trickstate_next);
		state.unplay();
		if ((thismask & equivalents[i]) != 0) thismask |= sameRankOrHigher[moves[i]];
		if (thisPlayWorks) {
			if (winmove) *winmove = state.cardsPlayed[state.nCardsPlayed];
			rwmask |= thismask;
			return true;
		} else {
			failmask |= thismask;
		}
	}
	
	// If we get here, we didn't find a winning move
	rwmask |= failmask;
	return false;
}


// This is the search function for the last player to play to the trick
bool analyzer::search_pl3(uint tricktarget, player_t pl, uint64& rwmask, card_t* winmove, const rankequiv_t& rankequiv, const trickstate_t& trickstate)
{
	// Enumerate possible moves
	card_t moves[13];
	uint64 equivalents[13];
	int movecount = state.generateMoves(pl, trickstate, rankequiv, moves, equivalents);
	
	// Prepare for searching
	int oppotarget = 1 + state.tricksLeft() - tricktarget;	
	uint64 failmask = 0;					// winning cards in failing lines
	
	// Update rank equivalence
	rankequiv_t rankequiv_next = rankequiv;
	for (int ci = state.nCardsPlayed - 4; ci < state.nCardsPlayed; ci++) rankequiv_next.play(state.cardsPlayed[ci]);
	
	// Try each move
	for (int i = 0; i < movecount; i++) {
		bool thisPlayWorks = false;
		uint64 thismask = 0;
		state.play(moves[i], pl);
		rankequiv_next.play(moves[i]);
		trickstate_t trickstate_this = trickstate;
		trickstate_this.play(pl, moves[i], deal.trumps);
		if (partnership(pl) == partnership(trickstate_this.winner)) {
			thisPlayWorks = search_pl0(tricktarget - 1, trickstate_this.winner, thismask, 0, rankequiv_next);
		} else {
			thisPlayWorks = !search_pl0(oppotarget - 1, trickstate_this.winner, thismask, 0, rankequiv_next);
		}
		if (trickstate_this.ranktrick) thismask |= sameRankOrHigher[trickstate_this.winningcard];
		if ((thismask & equivalents[i]) != 0) thismask |= sameRankOrHigher[moves[i]];
		state.unplay();
		rankequiv_next.unplay(moves[i]);
		if (thisPlayWorks) {
			if (winmove) *winmove = state.cardsPlayed[state.nCardsPlayed];
			rwmask |= thismask;
			return true;
		} else {
			failmask |= thismask;
		}
	}
	
	// If we get here, we didn't find a winning move
	rwmask |= failmask;
	return false;
}

// This is the search function; for a generic player - we now simply delegate to the relevant search_pl<n> function
bool analyzer::search(uint tricktarget, player_t pl, card_t* winmove)
{
	// Facts about the deal so far
	rankequiv_t rankequiv;
	uint64 rwmask = 0;
	
	// Start of a trick
	if (state.startOfTrick()) return search_pl0(tricktarget, pl, rwmask, winmove, rankequiv);
	
	// Middle of a trick, so must do some bookeeping
	int istart = state.nCardsPlayed & ~3;
	trickstate_t trickstate(state.whoPlayed[istart], state.cardsPlayed[istart]);
	for (int j = istart+1; j < state.nCardsPlayed; j++) trickstate.play(state.whoPlayed[j], state.cardsPlayed[j], deal.trumps);
	
	// Delegate to apppropriate search_pl<n> method
	switch (state.nCardsPlayed % 4) {
		case 0: return search_pl0(tricktarget, pl, rwmask, winmove, rankequiv);
		case 1: return search_pl1(tricktarget, pl, rwmask, winmove, rankequiv, trickstate);
		case 2: return search_pl2(tricktarget, pl, rwmask, winmove, rankequiv, trickstate);
		case 3: return search_pl3(tricktarget, pl, rwmask, winmove, rankequiv, trickstate);
	}
}

// Constructor
analyzer::analyzer(DealController_Analyzer* dc, const deal_t& d, cache_t* cache) :
	dc(dc), state(d), deal(d), cache(cache), plnext_ui(nextpl(d.declarer))
{
}

// Zero-Window Search Function
int analyzer::sureTricks(player_t pl, card_t* winmove) 
{
	int low = 0, high = 1 + state.tricksLeft();
	while (low+1 < high) {
		int goal = (low + high) / 2;
		bool rv = search(goal, pl, winmove);
		if (rv) low = goal; else high = goal;
	}
	return low;
}	

// Analysis function
void analyzer::analyze(bool analyzeAllOptions) 
{
	// Find out whose go it is
	player_t pl = plnext_ui;
	
	// Find how many tricks are available from the current state
	card_t winmove = -1;
	int suretricks = sureTricks(pl, &winmove);
	std::cout << suretricks << " tricks available\n"; 
	if (winmove != -1) std::cout << "Best play: " << cardtext(winmove) << std::endl;
		
	// Generate list of possible moves from this state
	card_t moves[13];
	uint64 equivalents[13];
	if (state.startOfTrick()) trickstate_ui.ledsuit = nt;
	int movecount = state.generateMoves(pl, trickstate_ui, rankequiv_ui, moves, equivalents);
	
	// Record number of tricks for this move where known
	int movetricks[13];

	// Special case - if overall analysis is zero tricks, then all cards are the same
	if (suretricks == 0) {
		for (int i = 0; i < movecount; i++) movetricks[i] = 0;
	} else {
		for (int i = 0; i < movecount; i++) {
			if (moves[i] == winmove) movetricks[i] = suretricks;
			else movetricks[i] = -1;
		}
	}

	// Tell the controller about each move we know about
	for (int i = 0; i < movecount; i++)
		if (movetricks[i] != -1) {
			uint64 cardmask = equivalents[i];
			while (cardmask) {
				uint64 mcard = lsb(cardmask);
				dc->setSureTricks(bitindex(mcard), movetricks[i]);
				cardmask ^= mcard;
			}
		}
	
	// If we're not analyzing alternatives, we're done
	if (!analyzeAllOptions) return;
	
	// Analyse the unknown moves, if this is the last card of a trick
	if (state.nCardsPlayed % 4 == 3) {
		
		// For each move
		for (int i = 0; i < movecount; i++) {
			if (movetricks[i] == -1) {
				
				// Assemble trick state so far
				int istart = state.nCardsPlayed & ~3;
				trickstate_t trickstate(state.whoPlayed[istart], state.cardsPlayed[istart]);
				for (int j = istart+1; j < state.nCardsPlayed; j++) trickstate.play(state.whoPlayed[j], state.cardsPlayed[j], deal.trumps);
				
				// Update the position with this move
				state.play(moves[i], pl);
				trickstate.play(pl, moves[i], deal.trumps); 

				// Compute sure tricks (for the next player) from the subsequent position
				if (partnership(pl) == partnership(trickstate.winner)) {
					movetricks[i] = 1 + sureTricks(trickstate.winner);
				} else {
					movetricks[i] = state.tricksLeft() - sureTricks(trickstate.winner);
				}
				
				// Apply the result to all equivalent moves
				uint64 cardmask = equivalents[i];
				while (cardmask) {
					uint64 mcard = lsb(cardmask);
					dc->setSureTricks(bitindex(mcard), movetricks[i]);
					cardmask ^= mcard;
				}
				
				// Restore the position
				state.unplay();
			}
		}
	}
		
	// Analyse the unknown moves, if this is not the end of a trick
	else {
		for (int i = 0; i < movecount; i++) {
			if (movetricks[i] == -1) {
				
				// Update the position with this move
				state.play(moves[i], pl);
		
				// Compute sure tricks (for the next player) from the subsequent position
				player_t plnext = nextpl(pl);
				movetricks[i] = state.tricksLeft() - sureTricks(plnext);

				// Apply the result to all equivalent moves
				uint64 cardmask = equivalents[i];
				while (cardmask) {
					uint64 mcard = lsb(cardmask);
					dc->setSureTricks(bitindex(mcard), movetricks[i]);
					cardmask ^= mcard;
				}

				// Restore the position
				state.unplay();
			}
		}
	}
}

// Play a card - this is the UI updating us on the play so far; these cards cannot be unplayed
bool analyzer::playCard(card_t card)
{
	if (state.nCardsPlayed % 4 == 0) trickstate_ui = trickstate_t(plnext_ui, card); else trickstate_ui.play(plnext_ui, card, deal.trumps);
	state.play(card, plnext_ui);
	if (state.nCardsPlayed % 4 == 0) for (int ci = state.nCardsPlayed - 4; ci < state.nCardsPlayed; ci++) rankequiv_ui.play(state.cardsPlayed[ci]);
	if (state.nCardsPlayed % 4 == 0) plnext_ui = trickstate_ui.winner; else plnext_ui = nextpl(plnext_ui);
	return true;
}
