// This file is part of FreeFinesse, a double-dummy analyzer (c) Ed Colley, 2006
// It is made available under the GPL; see the file COPYING for details

//
//  Implementation of the deal controller
//

#include "stdafx.h"

// Implementation of Controller
class DealController_Impl : public DealController_UI, DealController_Analyzer {
public:
	DealController_Impl();
	virtual bool setDeal(const deal_t&);
	virtual bool clickCard(card_t);
	virtual void addView(DealView*);
	virtual void removeView(DealView*);
	virtual void updateView(DealView*);
	virtual void setSureTricks(card_t, int tricks);
	virtual void analyze(bool analyzeAllOptions);
private:
	void invalidateAnalysis();
	std::vector<DealView*> views;
	deal_t deal;					// the deal
	CardState state[52];			// state of every card
	int nCardsPlayed;				// number of cards played so far in total
	card_t cardsPlayed[52];			// what they were
	analyzer* analyzer_;			// analyzer for the current position; also keeps track of whose go it is
	cache_t cache;					// cache of analysis results; we invalidate this when the deal changes
	void setState(card_t, CardState::cs_state);
};

// Factory function
DealController_UI* createDealController() 
{
	return new DealController_Impl;
}

// A card has been clicked
bool DealController_Impl::clickCard(card_t c) 
{	
	// Check that this card is part of the deal
	if (state[c].state == CardState::cs_notdealt) return false;
	
	// Taking back a card already played
	if (state[c].state == CardState::cs_played) {
		int prevCardsPlayed = nCardsPlayed;
		for (int i = 0; i < nCardsPlayed; i++)
			if (cardsPlayed[i] == c) { nCardsPlayed = i; break; }
		for (int i = nCardsPlayed; i < prevCardsPlayed; i++)
			setState(cardsPlayed[i], CardState::cs_unplayed);
		invalidateAnalysis();
		return true;
	}
	
	// Playing a card - TODO check this is a valid play
	// TODO - allow clicking on alternative choice for other players to backup & restart analysis
	if (state[c].state == CardState::cs_unplayed) {
		if (state[c].player != analyzer_-> plnext_ui) return false;
		setState(c, CardState::cs_played);
		cardsPlayed[nCardsPlayed++] = c;
		invalidateAnalysis();
		return true;
	}
	
	// Invalid play
	return false;
}

// Add view
void DealController_Impl::addView(DealView* v)
{
	views.push_back(v);
	updateView(v);
}


// Remove view
void DealController_Impl::removeView(DealView* v) 
{
	for (std::vector<DealView*>::iterator view = views.begin(); view != views.end(); view++) {
		if (v == *view) { views.erase(view); return; }
	}
}


// Update a (new) view with the current state
void DealController_Impl::updateView(DealView* view) 
{
	for (card_t c = 0; c < 52; c++) {
		view->setState(c, state[c]);
	}
}


// Create analyzer; we do this from scratch every time something changes rather than attempt to keep everything in synch
void DealController_Impl::invalidateAnalysis() 
{
	for (int c = 0; c < 52; c++) if (state[c].sureTricks != -1) setSureTricks(c, -1);
	delete analyzer_;
	analyzer_ = new analyzer(this, deal, &cache);
	for (int i = 0; i < nCardsPlayed; i++) analyzer_->playCard(cardsPlayed[i]);
}


// Perform analysis
void DealController_Impl::analyze(bool analyzeAllOptions) 
{
	analyzer_->analyze(analyzeAllOptions);
}


// A whole new deal
bool DealController_Impl::setDeal(const deal_t& d) 
{
	deal = d;
	cache.clear();
	for (int pl = 0; pl < 4; pl++) {
		for (int s = 0; s < 4; s++) {
			for (std::vector<card_t>::const_iterator it = deal.cards[pl][s].begin(); it != deal.cards[pl][s].end(); it++) {
				state[*it].player = player_t(pl);
				state[*it].state = CardState::cs_unplayed;
				state[*it].sureTricks = -1;
			}
		}
	}
	nCardsPlayed = 0;
	invalidateAnalysis();
	for (std::vector<DealView*>::iterator view = views.begin(); view != views.end(); view++) {
		updateView(*view);
	}
	return true;
}


// Constructor
DealController_Impl::DealController_Impl() : nCardsPlayed(0), analyzer_(0) 
{
	for (int i = 0; i < 52; i++) { 
		state[i].player = player_t(-1);
		state[i].state = CardState::cs_notdealt;
		state[i].sureTricks = -1;
	}		
}


// Set the state of a card, and broadcast it
void DealController_Impl::setState(card_t c, CardState::cs_state cs)
{
	state[c].state = cs;
	for (std::vector<DealView*>::iterator view = views.begin(); view != views.end(); view++) {
		(*view)->setState(c, state[c]);
	}
}


// Set the sure number of tricks
void DealController_Impl::setSureTricks(card_t c, int tricks)
{
	state[c].sureTricks = tricks;
	for (std::vector<DealView*>::iterator view = views.begin(); view != views.end(); view++) {
		(*view)->setState(c, state[c]);
	}
}
