// This file is part of FreeFinesse, a double-dummy analyzer (c) Ed Colley, 2006
// It is made available under the GPL; see the file COPYING for details

//
//  Keeping track of the current game state for analysis purposes
//

#include "stdafx.h"

// Generate all possible unique moves for a player (not the first to play to the trick)
int gamestate_t::generateMoves(player_t pl, const trickstate_t& trickstate, const rankequiv_t& rankequiv, card_t* moves, uint64* equivalents) const
{
	// Kep count of the moves
	int movecount = 0;
	
	// If we can follow suit, then do that
	uint64 suit = mPlayerHand[pl] & suitmask[trickstate.ledsuit];
	if (suit) {
		
		// Generate unique moves (lowest card first)
		card_t suitmoves[13];
		uint64 suitequivalents[13];
		card_t lastmove = -1;
		while (suit) {
			uint64 mcard = lsb(suit);
			suit ^= mcard;
			card_t thismove = bitindex(mcard);
			if (rankequiv.nextlower[thismove] != lastmove) {
				suitmoves[movecount] = thismove;
				suitequivalents[movecount] = mcard;
				movecount++;
			} else {
				suitequivalents[movecount-1] |= mcard;
			}
			lastmove = thismove;
		}
		
		// Order moves - first highest...
		moves[0] = suitmoves[movecount-1];
		equivalents[0] = suitequivalents[movecount-1];
		
		// ...then lowest...
		if (movecount > 1) {
			moves[1] = suitmoves[0];
			equivalents[1] = suitequivalents[0];
		}
		
		// ..then top-down.
		for (int i = 2; i < movecount; i++) {
			moves[i] = suitmoves[movecount-i];
			equivalents[i] = suitequivalents[movecount-i];
		}
	}

	// Otherwise, consider other suits
	else {
		// Assemble unique moves in each suit (lowest first)
		card_t suitmoves[4][13];
		int suitmovecount[4];
		uint64 suitequivalents[4][13];
		for (int s = 0; s < 4; s++) {
			suitmovecount[s] = 0;
			card_t lastmove = -1;
			uint64 suit = mPlayerHand[pl] & suitmask[s];
			while (suit) {
				uint64 mcard = lsb(suit);
				suit ^= mcard;
				card_t thismove = bitindex(mcard);
				if (rankequiv.nextlower[thismove] != lastmove) {
					suitequivalents[s][suitmovecount[s]] = mcard;
					suitmoves[s][suitmovecount[s]] = thismove;
					suitmovecount[s]++;
				} else {
					suitequivalents[s][suitmovecount[s]-1] |= mcard;
				}
				lastmove = thismove;
			}
		}
		
		// First, a low ruff
		if ((deal.trumps != nt) && (suitmovecount[deal.trumps] > 0)) {
			moves[movecount] = suitmoves[deal.trumps][0];
			equivalents[movecount] = suitequivalents[deal.trumps][0];
			movecount++; 
		}
		
		// Then a low discard from each non-trump suit
		for (int s = 0; s < 4; s++) {
			if (s == deal.trumps) continue;
			if (suitmovecount[s] > 0) {
				moves[movecount] = suitmoves[s][0];
				equivalents[movecount] = suitequivalents[s][0];
				movecount++;
			}
		}
		
		// Then higher ruffs
		if (deal.trumps != nt)
			for (int i = 1; i < suitmovecount[deal.trumps]; i++) {
				moves[movecount] = suitmoves[deal.trumps][i];
				equivalents[movecount] = suitequivalents[deal.trumps][i];
				movecount++;
			}
		
		// Then higher discards
		for (int s = 0; s < 4; s++) {
			if (s == deal.trumps) continue;
			for (int i = 1; i < suitmovecount[s]; i++) {
				moves[movecount] = suitmoves[s][i];
				equivalents[movecount] = suitequivalents[s][i];
				movecount++;
			}
		}
	}
	
	// Return the number of moves
	return movecount;
}


// Who holds a particular card (from the deal)
player_t deal_t::holder(card_t c) const
{
	suit_t s = suit(c);
	for (int pl = 0; pl < 4; pl++) {
		for (std::vector<card_t>::const_iterator it = cards[pl][s].begin(); it != cards[pl][s].end(); it++) {
			if (*it == c) return player_t(pl);
		}
	}
	throw "Card not held by anyone";
}


// Initialise a deal
void deal_t::setCards(player_t pl, const std::vector<card_t>& cardvec) 
{
	cardsPerPlayer = 0;
	for (int s = 0; s < 4; s++) cards[pl][s].clear();
	for (std::vector<card_t>::const_iterator it = cardvec.begin(); it != cardvec.end(); it++) {
		cards[pl][suit(*it)].push_back(*it);
		cardsPerPlayer++;
	}
}


// Construct the game state
gamestate_t::gamestate_t(const deal_t& deal) : deal(deal), nCardsPlayed(0), mCardsLeft(0), uSuitLengths(0)
{
	for (int pl = 0; pl < 4; pl++) {
		mPlayerHand[pl] = 0;
		for (int s = 0; s < 4; s++) {
			std::vector<card_t>::const_iterator it;
			for (it = deal.cards[pl][s].begin(); it != deal.cards[pl][s].end(); it++) {
				uSuitLengths += uint64(1) << 4*(4*pl+s);
				mCardsLeft |= bit(*it);
				mPlayerHand[pl] |= bit(*it);
			}
		}
	}	
}

// Output operator (for debugging)
std::ostream& operator<<(std::ostream& out, const gamestate_t& state)
{
	for (int pl = 0; pl < 4; pl++) {
		if (pl != 0) out << " ";
		out << playertext(player_t(pl)) << "[";
		for (int s = 0; s < 4; s++) {
			const std::vector<card_t>& c = state.deal.cards[pl][s];
			for (std::vector<card_t>::const_iterator it = c.begin(); it != c.end(); it++) {
				if (state.mCardsLeft & (uint64(1) << *it)) out << ranktext(*it);
			}
			out << '.';
		}
		out << "]";
	}
	return out;
}
