// This file is part of FreeFinesse, a double-dummy analyzer (c) Ed Colley, 2006
// It is made available under the GPL; see the file COPYING for details

//
//  Structures for the current state of the game for the analyzer
//

#include <ostream>

// Definition of a deal
struct deal_t {
	std::string description;
	std::vector<std::string> playernames;
	player_t declarer;
	std::vector<card_t> cards[4][4];		// [player][suit]
	suit_t trumps;
	int cardsPerPlayer;
	player_t holder(card_t) const;
	void setCards(player_t, const std::vector<card_t>&);
};

// Rank equivalence (kept dynamically up-to-date)
struct rankequiv_t {
	card_t nexthigher[52];
	card_t nextlower[52];
	rankequiv_t() {
		for (int s = 0; s < 4; s++) {
			nextlower[0 * 4 + s] = 0 * 4 + s;		// deuce
			nexthigher[0 * 4 + s] = 1 * 4 + s;		// deuce
			for (int r = 1; r < 12; r++) {
				nextlower[r * 4 + s] = (r-1) * 4 + s;
				nexthigher[r * 4 + s] = (r+1) * 4 + s;
			}
			nextlower[12 * 4 + s] = 11 * 4 + s;		// Ace
			nexthigher[12 * 4 + s] = 12 * 4 + s;	// Ace
		}
	}
	void play(card_t c) {
		card_t nh = nexthigher[c];
		card_t nl = nextlower[c];
		nexthigher[nl] = nh;
		nextlower[nh] = nl;
	}
	void unplay(card_t c) {				// can't unplay more than one card
		card_t nh = nexthigher[c];
		card_t nl = nextlower[c];
		if (nl != c) nexthigher[nl] = c;
		if (nh != c) nextlower[nh] = c;
	}
};


// Progress of current trick
struct trickstate_t {
	player_t leader;
	player_t winner;
	card_t winningcard;
	bool ranktrick;
	suit_t ledsuit;
	suit_t winsuit;
	trickstate_t(player_t leader, card_t card) : leader(leader), winner(leader),
		winningcard(card), ranktrick(false), ledsuit(suit(card)), winsuit(suit(card)) { }
	trickstate_t() { }
	
	// Record a play
	void play(player_t pl, card_t card, suit_t trumps) {
		suit_t s = suit(card);
		if (s == winsuit) {
			ranktrick = true;
			if (card > winningcard) {
				winningcard = card;
				winner = pl;
			}
		} else if (s == trumps) {
			ranktrick = false;
			winningcard = card;
			winsuit = trumps;
			winner = pl;
		}
	}
};

// State of the game
struct gamestate_t {

	// Construction
	gamestate_t(const deal_t&);
	const deal_t& deal;

	// The play so far
	int nCardsPlayed;				// number of cards played so far in total
	card_t cardsPlayed[52];			// what they were
	player_t whoPlayed[52];			// who played each card

	// Derived info about the state of play
	uint tricksLeft() const { return deal.cardsPerPlayer-nCardsPlayed/4; }
	bool startOfTrick() const { return nCardsPlayed%4 == 0; }
	
	// Current state of play
	uint64 mCardsLeft;				// mask cards left in total (deck order)
	uint64 mPlayerHand[4];			// mask cards left per player (deck order)
	uint64 uSuitLengths;			// number of cards left per player per suit

	// Play a card
	void play(card_t c, player_t pl) {
		cardsPlayed[nCardsPlayed] = c;
		whoPlayed[nCardsPlayed] = pl;
		nCardsPlayed++;
		mPlayerHand[pl] ^= bit(c); 
		mCardsLeft ^= bit(c); 
		uSuitLengths -= uint64(1) << 4*(4*pl+suit(c));
	}

	// Take back a play
	void unplay() {
		nCardsPlayed--;
		card_t c = cardsPlayed[nCardsPlayed];
		player_t pl = whoPlayed[nCardsPlayed];
		mPlayerHand[pl] ^= bit(c);
		mCardsLeft ^= bit(c);
		uSuitLengths += uint64(1) << 4*(4*pl+suit(c));
	}

	// Generate a list of legal moves, with duplicates eliminated (but reported seperately)
	int generateMoves(player_t, const trickstate_t&, const rankequiv_t&, card_t* moves, uint64* equivalents) const;
};


std::ostream& operator<<(std::ostream& out, const gamestate_t& state);
